function [x_tot,x,x_t]=Upper_Layer_ParMH(N,T,sigprop,tp)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%   IMPLEMENTATION OF MULTIPLE IMPORTANCE SAMPLING SCHEMES   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%   WITHIN LAYERED ADAPTIVE IMPORTANCE SAMPLING (LAIS)   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%      (see references below)                     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% V. Elvira, L. Martino, D. Luengo, M. F. Bugallo,
%%% "Generalized Multiple Importance Sampling",
%%% https://arxiv.org/abs/1511.03095
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% L. Martino, V. Elvira, D. Luengo, J. Corander,
%%% "Layered Adaptive Importance Sampling", Statistics and Computing, 2016.
%%% doi:10.1007/s11222-016-9642-5
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

dim = tp.dim;
if N<2
    N=2;
end
if T<1
    T=1;
end
disp('-----------------------------------------------------------------------------------------')
disp('****-****-****-****-****-****-****-****')
disp('****       UPPER-LAYER            *****')
disp('****-****-****-****-****-****-****-****')
disp('-----------------------------------------------------------------------------------------')
disp(['Number of chains= ' num2str(N), ' '])
disp(['Iterations= ' num2str(T)])
disp(['Total number of generated mean parameters= ' num2str(N*T)])
disp('-----------------------------------------------------------------------------------------')

if tp.type == 7 % banana-shaped target
    logf=@(x) target(x.',tp); % transpose
    initialPoints=-5+10*rand(N,dim);
    x{1}=initialPoints;
    SIGMA=sigprop.^2*eye(dim);
end

x_tot=[];

for nn=1:N
    Vbef(1,nn)=logf(x{1}(nn,:));
end

%% Simulation

for t=2:T+1
    x{t}=mvnrnd(x{t-1},SIGMA);
    for nn=1:N
        Vnow(1,nn)=logf(x{t}(nn,:));
    end
    rho=exp(Vnow-Vbef);
    alpha=min([ones(1,N);rho]);
    u=rand(1,N);
    test=(u<=alpha);
    vec_aux=(x{t}-x{t-1}).*repmat(test',1,dim);
    vec_aux2=(Vnow-Vbef).*test;
    x{t}=x{t-1}+vec_aux;
    Vbef=Vbef+vec_aux2;
    x_tot=[x_tot x{t}'];
end %%% end for - t
x{1}=[];
x=x(~cellfun('isempty',x));
for i=1:N
    x_t{i}=x_tot(:,i:N:end);
end







